# FILE: mmdisp.tcl
#
# Main Tcl/Tk script for OOMMF 2D display program
#
# Requires Tcl/Tk script files: misc.tcl dialog.tcl
#  They must be source'd before this one, in that order.
#

set omfmmd_library [file dirname [Oc_DirectPathname [info script]]]
package require Oc
package require Ow
package require Vf

# Disable Oc's user-invoked auto-size feature---mmDisp handles its own
# resizing.
Oc_DisableAutoSize .

########################### PROGRAM DOCUMENTATION ###################
Oc_Main SetAppName mmDisp
Oc_Main SetVersion 1.0.1.0
regexp \\\044Date:(.*)\\\044 {$Date: 1998/10/03 00:28:28 $} _ date
Oc_Main SetDate [string trim $date]
regexp \\\044Author:(.*)\\\044 {$Author: donahue $} _ author
Oc_Main SetAuthor [Oc_Person Lookup [string trim $author]]
Oc_Main SetExtraInfo "Built against Tcl $omfTclBuildVersion\n\
	              Running on Tcl [info patchlevel]"
Oc_Main SetHelpURL [Oc_Url FromFilename [file join [file dirname \
        [file dirname [file dirname [Oc_DirectPathname [info \
        script]]]]] doc userguide Vector_Field_Display_mmDisp.html]]
wm title . [Oc_Main GetAppName]<[pid]>  ;# Window title to display while
					## program is initializing.

# Global arrays: mmdisp, _mmdisp

set mmd_log [em_extension ApplicationLog]

# If the --server option appeared the command line, start a server.
if {$mmd_serverflag} {
    package require Net
    Net_Protocol New mmd_protocol -name [list OOMMF vectorField protocol 0.0]
    $mmd_protocol AddMessage start bye {} {
        return [list close [list 0 Bye!]]
    }
    $mmd_protocol AddMessage start datafile { fnlist } {
	# CAUTION!!! -- The command [AsyncTempFileDisplayRequest]
	# re-enters the event loop, so when an instance of Net_Connection
	# calls this proc via a call to [$net_protocol Reply], it can
	# be re-entered, and do such nasty things as destroy itself.
	AsyncTempFileDisplayRequest $fnlist
	return [list start [list 0]]
    }
    $mmd_protocol AddMessage start HasGui {} {
        return [list start [list 0 0]]
    }
    Net_Server New mmd_server -protocol $mmd_protocol \
            -log [em_extension ApplicationLog] -register 1 \
            -alias [Oc_Main GetAppName]
    # Start server on an OS-selected port.
    $mmd_server Start 0

    # Set up to be sure that server shuts down on exit.  
    rename exit tcl_exit
    proc exit {{code 0}} {
        global mmd_server
        if {[info exists mmd_server]} {
            catch { $mmd_server Delete }  ;# Safety catch
        }
        # Do something smarter eventually.  For now, 2 seconds should
        # be enough time for the server to shut down cleanly.
        after 2000 tcl_exit $code
    }
}

set _mmdisp(atfdr_filelist) {}   ;# File display request stack
set _mmdisp(atfdr_procsem) 0     ;# Processing semaphore
proc AsyncTempFileDisplayRequest { fnlist } {
    global mmdisp _mmdisp

    # Append fnlist to request stack
    lappend _mmdisp(atfdr_filelist) $fnlist

    # If semaphore already taken, just return
    if {$_mmdisp(atfdr_procsem)} return

    # Otherwise, schedule delete of all files on list except last,
    # which is displayed and then deleted.  After this check the
    # list for new names, which may have been added during display
    # processing.
    set _mmdisp(atfdr_procsem) 1     ;# Grab semaphore, and
    ## wrap rest up in a catch so we are sure to unset semaphore
    catch {
	update
	while {[llength $_mmdisp(atfdr_filelist)]>0} {
	    set worklist $_mmdisp(atfdr_filelist)
	    set _mmdisp(atfdr_filelist) {}  ;# Clear stack
	    
	    set namepair [lindex $worklist end]
	    set filename [lindex $namepair 0]
	    set title    [lindex $namepair 1]

	    # Display file
	    if {[string match "Vf_EmptyMesh" [GetMeshType]]} {
		# If no file currently displayed, use auto-display
		ReadFile $filename $title
	    } else {
		# Otherwise, keep all current size and display parameters
		global ViewportWidth ViewportHeight DisplayRotation
		set autodisplay(sample) 0
		set autodisplay(zoom) 0
		ReadFile $filename $title $ViewportWidth $ViewportHeight \
			$DisplayRotation autodisplay
	    }

	    # Delete all files on stack
	    foreach namepair $worklist {
		catch {file delete [lindex $namepair 0]}
	    }

	    # Loop back and check list for more files
	    update
	}
    }  ;# End of semaphore-set catch
    set _mmdisp(atfdr_procsem) 0     ;# Release semaphore
}

# Allow loading of other support procs
lappend auto_path [file dirname [Oc_DirectPathname [info script]]]

source [file join [file dirname [Oc_DirectPathname [info script]]] misc.tcl]

########################### GLOBAL VARIABLES ########################
# It would probably be better to have many of these sitting inside the
# _mmdisp array variable.
if {![info exists DEBUGLVL]} { set DEBUGLVL 0 } ;# Value of C++ compile-
                      ## time macro DEBUGLVL.  Should be set inside
                      ## Oommf_Init() (currently in OOMMF/oomf/oomfInit.cc)
set DrawFrameCount  0 ;# This variable is incremented by DrawFrame each
                      ## time it is called, so this variable can be used
                      ## for profiling.
set UpdateViewCount 0 ;# Any time the screen display is modified, this
                      ## value will be incremented.  A trace on this
                      ## variable can be used by any routine (e.g. the
                      ## print dialog) that needs to take action upon
                      ## such events.  Currently this variable is written
                      ## to only by the PackViewport proc.  NOTE: Scrolling
                      ## changes are NOT tracked by this variable, at this
                      ## time. (-mjd 97/2/11)
set canvas_frame .canvas_frame
set canvas $canvas_frame.canvas   ;# Name of canvas widget

set DefaultWindowWidth   640
set DefaultWindowHeight  480
set DefaultCellSize 25
# Note: DefaultCellSize must be > 0
set DefaultViewportWidth  $DefaultWindowWidth  ;# Safety.  These should be
set DefaultViewportHeight $DefaultWindowHeight ;# set during Init phase.
set MinViewportWidth  1
set MinViewportHeight 1
set ViewportWidth     1
set ViewportHeight    1
set CanvasWidth       1
set CanvasHeight      1

# NOTE: There are 3 different width/height pairs, and at any time
#   they may have completely different values.  ViewportWidth/Height
#   is requested viewport width/height, *including* scrollbars as needed.
#   The requested value will be modified as necessary to meet the
#   MinViewportWidth/Height restrictions.  The actual size of the viewing
#   area, *not* including the scrollbars, is stored in the CanvasWidth
#   and CanvasHeight variables.  The whole canvas size can be obtained
#   from the canvas "-scrollregion" configuration option, or from the
#   "frame" C++ class by calling GetFrameBox, or the Tcl procs "GetPlotBox"
#   and "GetPlotExtents".  The GetFrameBox returns a 6-tuple,
#                   xmin ymin zmin xmax ymax zmax
#   which is an overestimate of the box needed to enclose all drawn points.
#   GetPlotBox returns a 4-tuple, "xmin ymin xmax ymax", which is drawn
#   from the GetFrameBox result but further expanded by adding in the user
#   specified margin.  GetPlotExtents returns the width and height of the
#   box returned by GetPlotBox.
#
set FrameScale        0
set DisplayRotation   0
set FileId           0     ;# Incremented each time a new file is loaded.
                           ## Traces may be set on this variable to update
                           ## file specific quantities, e.g., filename,
                           ## file description, mesh type, etc.
set RealFileId       0     ;# Incremented each time a non-empty file is
                           ## read.
set FrameSubsampleIncrement 0.1  ;# Subsample slider increment size
set ColorMap(base) black   ;# Color used for in-plane arrows
set ColorMap(in)   red     ;# Color used for arrows pointing into-page
set ColorMap(out)  blue    ;# Color used for arrows pointing out-of-page

# Trough color for 'scale' widgets and 'scrollbar' widgets.
# By default, Windows sets these to same as default frame background.
scale .mmdtemp
set temp [winfo rgb . [.mmdtemp cget -background]]
set tempr [expr round([lindex $temp 0]*0.85)]
set tempg [expr round([lindex $temp 1]*0.85)]
set tempb [expr round([lindex $temp 2]*0.85)]
set _mmdisp(troughcolor) [format "#%04X%04X%04X" $tempr $tempg $tempb]
unset temp tempr tempg tempb
destroy .mmdtemp


# Plot configuration array
array set plot_config {
    colormaps  { Red-Black-Blue Blue-White-Red Teal-White-Red \
            Black-Gray-White White-Green-Black }
    quantitylist       {}
    arrow,status       1
    arrow,colormap     Red-Black-Blue
    arrow,colorcount   256
    arrow,quantity     z
    arrow,subsample    0
    arrow,mag          1
    pixel,status       0
    pixel,colormap     Red-Black-Blue
    pixel,colorcount   256
    pixel,quantity     x
    pixel,subsample    0
    pixel,mag          1
    misc,background    white
    misc,drawboundary  1
    misc,margin        10
    misc,confine       1
}

# Misc. constants
set SmallZoom 1.1486984  ;# Amt to enlarge on Page_Up (Prior). This is 2^(1/5)
set LargeZoom 2          ;# Amount to enlarge on Shift-Page_Up

# These scroll state variables should be modified *only* in
# proc PackViewport
set XScrollState      0
set YScrollState      0

# Whether or not the control bar is displayed
set omfCtrlBarState 1   ;# 1->on, 0->off

# Display state recover stack
# (Currently only 1 entry deep, but may be made deeper in future
set omfSavedDisplayState(zoom)     $FrameScale
set omfSavedDisplayState(xcenter)  0.5
set omfSavedDisplayState(ycenter)  0.5
set omfSavedDisplayState(rot)      $DisplayRotation
proc OMF_SaveDisplayState { {zoom ""} {xcenter ""}  {ycenter ""} {rot ""}} {
    global omfSavedDisplayState FrameScale DisplayRotation
    if {"$zoom"==""} { set zoom $FrameScale }
    if {"$xcenter"=="" || "$ycenter"==""} {
        GetScrollCenter curx cury
        if {"$xcenter"==""} { set xcenter $curx }
        if {"$ycenter"==""} { set ycenter $cury }
    }
    if {"$rot"==""} { set rot $DisplayRotation }
    set omfSavedDisplayState(zoom) $zoom
    set omfSavedDisplayState(xcenter)  $xcenter
    set omfSavedDisplayState(ycenter)  $ycenter
    set omfSavedDisplayState(rot)      $rot
}
proc OMF_RecoverDisplayState {} {
    global omfSavedDisplayState DisplayRotation
    # Note: SetScale calls OMF_SaveDisplayState
    #   & thereby reset these values
    set oldzoom     $omfSavedDisplayState(zoom)
    set oldxcenter  $omfSavedDisplayState(xcenter)
    set oldycenter  $omfSavedDisplayState(ycenter)
    set oldrot      $omfSavedDisplayState(rot)
    # Make rotation adjustments to xcenter & ycenter
    set dummywidth 0; set dummyheight 0
    OMF_RotateDimensions [expr $DisplayRotation-$oldrot] \
            dummywidth dummyheight oldxcenter oldycenter
    # Undo
    SetScale $oldzoom
    SetScrollCenter $oldxcenter $oldycenter
} 

##### Window geometry requirements #####################################
set RootHeadHeight 0           ;# Height of menu + action bar
set ScrollCrossDim 0           ;# Width of yscroll, height of xscroll

proc OMF_SetGeometryRequirements {} {
    global MinViewportWidth MinViewportHeight RootHeadHeight
    global omfCtrlBarState menubar ctrlbar

    if {$omfCtrlBarState} {
	WaitVisible $ctrlbar 500
	set size [expr [winfo reqheight $ctrlbar.subsample]\
		+[winfo reqheight $ctrlbar.framescale]]
	$ctrlbar.rotation configure -height $size -width $size
    }

    set MinViewportHeight 1
    set MinViewportWidth [winfo reqwidth $menubar]
    if {[string compare $menubar [winfo toplevel $menubar]]==0} {
       # New menu scheme in effect --> menu doesn't affect
       # positioning of "." subwindows.
       set RootHeadHeight 0
    } else {
       # Old menu scheme being used.  Adjust vertical offset
       # to allow space for menu.
       set RootHeadHeight [winfo reqheight $menubar]
    }
    if {$omfCtrlBarState} {
        set RootHeadHeight [expr $RootHeadHeight \
                + [winfo reqheight $ctrlbar]]
        set temp [winfo reqwidth $ctrlbar]
        if {$temp>$MinViewportWidth} {
            set MinViewportWidth $temp
        }
    }
    set minwidth $MinViewportWidth
    set minheight [expr $RootHeadHeight + $MinViewportHeight]
    wm minsize . $minwidth $minheight
    wm resizable . 1 1     ;# Safety.  This is supposed to be the default.
}

##### Combined interface to mesh object ################################
proc OMF_MeshInfo { command option } {
    # Currently only supports "cget" command, but in the future
    # "configure" should be added.
    if { ![string match "cget" "$command"] } {
	Ow_NonFatalError "Error in OMF_MeshInfo: Invalid command $cmd"
	return
    }
    set value {}
    switch -exact -- $option {
	-filename { set value [GetMeshName]  }
	-title    { set value [GetMeshTitle] }
	-gridtype { set value [GetMeshType]  }
	-description { set value [GetMeshDescription] }
	default   {
	    Ow_NonFatalError "Error in OMF_MeshInfo: Invalid option $option"
	    return
	}
    }
    return $value
}

########################## DEBUGGING PROC'S ############################
proc PrintDimensions { window } {
    set width [$window cget -width]
    if [catch { set height [$window cget -height] }] then {
        set height "-1"
    }
    global mmd_log
    $mmd_log Append [format "%25s  Width=%3d, Height=%3d" $window $width \
            $height]
}

proc PrintSizes {} {
    global mmd_log canvas_frame canvas
    global MinViewportWidth MinViewportHeight ViewportWidth ViewportHeight
    $mmd_log Append "MinViewportWidth/Height: $MinViewportWidth/$MinViewportHeight" 
    $mmd_log Append "   ViewportWidth/Height: $ViewportWidth/$ViewportHeight" 
    GetRootDimensions width height
    $mmd_log Append [format "%25s  Width=%3d, Height=%3d" . $width $height]
    PrintDimensions $canvas
    PrintDimensions $canvas_frame.xscroll
    PrintDimensions $canvas_frame.yscroll
    $mmd_log Append [$canvas configure -scrollregion]
}

proc PrintDrawFrameCount { name elt op } {
    global DrawFrameCount mmd_log
    $mmd_log Append "DrawFrame call count: $DrawFrameCount"
}
trace variable DrawFrameCount w PrintDrawFrameCount

proc PrintUpdateViewCount { name elt op } {
    global UpdateViewCount mmd_log
    $mmd_log Append "UpdateView call count: $UpdateViewCount"
}
trace variable UpdateViewCount w PrintUpdateViewCount

########################## DISPLAY PROC'S ###########################
proc UpdateWindowTitle {name elt op} {
    set title [file tail [GetMeshTitle]]
    if {"$title"==""} { 
        set title "[Oc_Main GetAppName]<[pid]>"
    } else {
        set title "[Oc_Main GetAppName]<[pid]>: $title"
    }
    wm title . $title
    wm iconname . $title
}
trace variable FileId w UpdateWindowTitle

proc GetCanvasBox {} {
    # Returns plot coordinates of canvas viewport
    global canvas CanvasWidth CanvasHeight
    return [list [$canvas canvasx 0] [$canvas canvasy 0] \
	    [$canvas canvasx $CanvasWidth] [$canvas canvasy $CanvasHeight]]
}

proc GetPlotBox {} {
    # Returns the actual drawing region {xmin ymin xmax ymax},
    # plus margin.  SPECIAL CASE: If xmin>=xmax and ymin>=ymax,
    # then the margin is not added.
    global canvas plot_config
    set bbox [GetFrameBox]  ;# Drawing region
    foreach { xmin ymin zmin xmax ymax zmax } $bbox {}
    set margin $plot_config(misc,margin)
    if { $xmin >= $xmax && $ymin >= $ymax } {
	return "$xmin $ymin $xmin $ymin"
    }
    set pbox [list \
	    [expr $xmin-$margin] \
	    [expr $ymin-$margin] \
	    [expr $xmax+$margin] \
	    [expr $ymax+$margin]]     ;# Add margin
    return $pbox
}

proc GetPlotExtents { {wname {}} {hname {}} } {
    set pbox [GetPlotBox]
    if {![string match {} $wname]} { upvar $wname width }
    if {![string match {} $hname]} { upvar $hname height }
    set width [expr [lindex $pbox 2] - [lindex $pbox 0]]
    set height [expr [lindex $pbox 3] - [lindex $pbox 1]]
    return [list $width $height]
}

proc GetScrollCenter { xcenter_name ycenter_name } {
    # Note: This code should work even if canvas "confine"
    #  is false.
    global canvas CanvasWidth CanvasHeight
    upvar $xcenter_name xcenter
    upvar $ycenter_name ycenter
    foreach {xmin ymin xmax ymax} [$canvas cget -scrollregion] {}
    set xdelta [expr $xmax-$xmin]
    if {$xdelta<=0.} {set xdelta 1.} ;# Safety
    set ydelta [expr $ymax-$ymin]
    if {$ydelta<=0.} {set ydelta 1.} ;# Safety

    set x [expr [$canvas canvasx 0] + $CanvasWidth/2.]
    ## Note: The input to the canvas canvasx (and canvasy) command
    ##  is a screen coordinate, in particular an integer.  The
    ##  command doesn't complain if you give in a non-integral value,
    ##  but the return value is integral (but with a decimal point).
    set xcenter [expr ($x-$xmin)/$xdelta]
    if {$xcenter<0.} {set xcenter 0}
    if {$xcenter>1.} {set xcenter 1.}

    set y [expr [$canvas canvasy 0] + $CanvasHeight/2.]
    set ycenter [expr ($y-$ymin)/$ydelta]
    if {$ycenter<0.} {set ycenter 0}
    if {$ycenter>1.} {set ycenter 1.}

    return [list $xcenter $ycenter]
}

proc SetScrollCenter { xcenter ycenter } {
    # All imports should be in range [0,1].  This code
    # should work even if canvas "confine" is turned off.
    global canvas CanvasWidth CanvasHeight
    foreach {xmin ymin xmax ymax} [$canvas cget -scrollregion] {}
    set xdelta [expr $xmax-$xmin]
    if {$xdelta<=0.} {set xdelta 1.} ;# Safety
    set ydelta [expr $ymax-$ymin]
    if {$ydelta<=0.} {set ydelta 1.} ;# Safety

    set xstart [expr $xcenter - $CanvasWidth/(2.*$xdelta)]
    set ystart [expr $ycenter - $CanvasHeight/(2.*$ydelta)]

    global plot_config
    if {0 && $plot_config(misc,confine)} {
	if {$xstart<0} {set xstart 0} elseif {$xstart>1} {set xstart 1}
	if {$ystart<0} {set ystart 0} elseif {$ystart>1} {set ystart 1}
    }
    $canvas xview moveto $xstart
    $canvas yview moveto $ystart
}

proc DetermineScrollStates { viewwidth viewheight framewidth frameheight \
        newxstatename newystatename } {
    # Determines which scrollbars will need to be activated, given a
    # canvas size of framewidth x frameheight, a viewport size of
    # viewwidth x viewheight, and a scrollbar cross dimension of
    # ScrollCrossDim.  States (0=>off, 1=>on) are returned in
    # newxstatename and newystatename.
    # Note: This code does *not* take MinViewportWidth/Height into account.
    global ScrollCrossDim
    upvar $newxstatename newxstate
    upvar $newystatename newystate
    set minheight [expr $viewheight - $ScrollCrossDim]
    set minwidth  [expr $viewwidth  - $ScrollCrossDim]
    if {($framewidth<=$viewwidth) && ($frameheight<=$viewheight)} {
        set newxstate 0; set newystate 0;
    } elseif {($framewidth>$viewwidth) && ($frameheight<=$minheight)} {
        set newxstate 1; set newystate 0;
    } elseif {($framewidth<=$minwidth) && ($frameheight>$viewheight)} {
        set newxstate 0; set newystate 1;
    } else {
        set newxstate 1; set newystate 1;
    }
}

proc PackViewport { {newviewwidth {}} {newviewheight {}} \
	{change_geometry 0}} {
    global CanvasWidth CanvasHeight
    global ViewportWidth ViewportHeight
    global MinViewportWidth MinViewportHeight
    global XScrollState YScrollState
    global ScrollCrossDim
    global canvas_frame canvas

    # Save current center for use in case scrollbars toggle on or off
    GetScrollCenter cx cy    

    # Set default values
    if {[string match {} $newviewwidth]} {
        set newviewwidth $ViewportWidth
    } else {
        set newviewwidth [expr round($newviewwidth)]
    }
    if {[string match {} $newviewheight]} {
        set newviewheight $ViewportHeight
    } else {
        set newviewheight [expr round($newviewheight)]
    }
    # Some Tk routines don't like window sizes smaller than 1
    if {$newviewwidth<1} {set newviewwidth 1}
    if {$newviewheight<1} {set newviewheight 1}

    set ViewportHeight $newviewheight
    set ViewportWidth $newviewwidth
    
    # Determine real (displayed) viewport width & height
    if { $newviewwidth < $MinViewportWidth } {
        set vwidth $MinViewportWidth
    } else {
        set vwidth $newviewwidth
    }
    if { $newviewheight < $MinViewportHeight } {
        set vheight $MinViewportHeight
    } else {
        set vheight $newviewheight
    }

    # Determine new scroll states & canvas display size
    set slop 1 ;# Amount of lap over to allow before requiring
               ## scrollbars.
    GetPlotExtents fwidth fheight
    DetermineScrollStates $vwidth $vheight \
            [expr $fwidth-$slop] [expr $fheight-$slop] \
            newxstate newystate
    if { $newystate } {
        set cwidth [expr $vwidth - $ScrollCrossDim]
    } else {
        set cwidth $vwidth
    }
    if { $newxstate } {
        set cheight [expr $vheight - $ScrollCrossDim]
    } else {
        set cheight $vheight
    }
    set cwidth [expr round($cwidth)]   ;# Canvas automatically rounds
    set cheight [expr round($cheight)] ;# to integers.

    # It appears that some (all) window managers won't allow
    # windows with dimensions less than 1.  Since we want the
    # canvas width and height to reflect that of the viewport,
    # adjust input values to meet this minimum size.
    if {$cwidth<1}  { set cwidth 1 }
    if {$cheight<1} { set cheight 1 }

    # Set new scrollregion
    $canvas configure -scrollregion [GetPlotBox]

    # Set canvas size
    set CanvasWidth $cwidth
    set CanvasHeight $cheight

    # If scrollstatus has changed, we need to repack display
    if { ($newxstate != $XScrollState) \
            || ($newystate != $YScrollState) } {
        # Unpack as needed
        if { $XScrollState && !$newxstate} {
            pack forget $canvas_frame.xscroll
	    set cx 0.5
        }
        if { $YScrollState } {
            # If $YScrollState is true, then y-scroll needs unpacking:
            # At least one of x-scroll or y-scroll is changing state,
            # and either way y-scroll packing changes.
            pack forget $canvas_frame.yscroll
            if { $XScrollState } {
                pack forget $canvas_frame.yscrollbox
            }
	    if {!$newystate} {
		set cy 0.5
	    }
        }
        # Now pack
        if { !$XScrollState && $newxstate } {
            pack $canvas_frame.xscroll -side bottom -fill x \
                    -before $canvas
        }
        if { $newystate } {
            # If $newystate is true, then y-scroll needs packing:
            # At least one of x-scroll or y-scroll is changing state,
            # and either way y-scroll packing changes.
            if { $newxstate } then {
                # Both x & y scrollbars are displayed; Use
                # corner box in lower righthand corner.
                pack $canvas_frame.yscroll -side right -fill y \
                        -in $canvas_frame.yscrollbox
                pack $canvas_frame.yscrollbox -side right -fill y \
                        -before $canvas_frame.xscroll
            } else {
                pack $canvas_frame.yscroll -side right -fill y \
                        -before $canvas
            }
        }
        set XScrollState $newxstate
        set YScrollState $newystate
	SetScrollCenter $cx $cy
    }

    if {$change_geometry} {
        # Change canvas geometry request.  NOTE: This *must* not
        # be activated from inside <Configure> event processing,
        # as it will likely generate a new <Configure> event, and
        # then the possibility exists for an infinite loop of
        # <Configure> events.  (Note: There appears to be less
	# redraw "bouncing" if 'wm geometry . {}' is placed _after_
	# the $canvas_frame resizing
        $canvas_frame configure -width $vwidth -height $vheight
        wm geometry . {}
    }

    global UpdateViewCount
    incr UpdateViewCount
}

###################### ACTION PROC'S ###############################
# ResizeViewport: Change viewport size.  This is bound to
# $canvas_frame <Configure>
proc ResizeViewport { width height } {
    # Skip if requested dimensions match current displayed dimensions.
    # (This helps protect stored sizes that are smaller than minimum
    # sizes.)
    global ViewportWidth MinViewportWidth
    global ViewportHeight MinViewportHeight
    if {$ViewportWidth < $MinViewportWidth } {
	set curviewwidth $MinViewportWidth
    } else {
	set curviewwidth $ViewportWidth
    }
    if {$ViewportHeight < $MinViewportHeight } {
	set curviewheight $MinViewportHeight
    } else {
	set curviewheight $ViewportHeight
    }
    if { $curviewwidth != $width || $curviewheight != $height } {
	PackViewport $width $height
	# Note that we must not call PackViewport from here with
	# change_geometry true, because that could generate an
	# infinite loop.
    }
}

proc DisplayRotationTrace { name elt op } {
    global DisplayRotation ViewportHeight ViewportWidth
    global canvas

    # Round new rotation direction to nearest multiple of 90
    set rot [expr 90*round($DisplayRotation/90.)]

    # Adjust to range [0,360)
    set rot [expr round(fmod($rot,360.))]
    if { $rot < 0 } { incr rot 360 }

    # Compare to previous rotation
    GetFrameRotation old_rotation
    set angle [expr $rot - $old_rotation]
    # Determine rotation step: 0, 1, 2 or 3
    set irotstep [expr round($angle/90.)%4]
    if { $irotstep == 0 } then {
        # No change
        GetFrameRotation DisplayRotation
        return
    }

    # Get current scroll positions
    GetScrollCenter xscrcenter yscrcenter

    # Rotate display
    SetFrameRotation $rot
    GetFrameRotation DisplayRotation
    UpdateCoordDisplay $rot

    # Addjust viewport width, height & scroll positions
    # Keep track of scrollbar centers rather than ends to
    # better handle case where displayed viewport width-height
    # values change due to MinViewportWidth/Height restrictions.
    set new_xscrcenter $xscrcenter
    set new_yscrcenter $yscrcenter
    set new_width      $ViewportWidth
    set new_height     $ViewportHeight
    OMF_RotateDimensions $angle new_width new_height \
            new_xscrcenter new_yscrcenter

    Ow_PushWatchCursor
    DrawFrame $canvas
    PackViewport $new_width $new_height 1
    SetScrollCenter $new_xscrcenter $new_yscrcenter
    Ow_PopWatchCursor
}

proc SetScale { scale } {
    global FrameScale canvas
    set scale [expr round($scale*100.)/100.]
    if { $scale == $FrameScale } { return }
    GetScrollCenter xcenter ycenter
    if { $scale > 0 } {
        OMF_SaveDisplayState $FrameScale $xcenter $ycenter
        SetFrameScale $scale
    }
    Ow_PushWatchCursor
    GetFrameScale FrameScale
    DrawFrame $canvas
    PackViewport
    global XScrollState YScrollState
    if {!$XScrollState} { set xcenter 0.5 }
    if {!$YScrollState} { set ycenter 0.5 }
    SetScrollCenter $xcenter $ycenter
    Ow_PopWatchCursor
}

proc RedrawDisplay {} {
    global canvas
    Ow_PushWatchCursor
    UpdatePlotConfiguration
    DrawFrame $canvas
    PackViewport
    Ow_PopWatchCursor
}

proc WrapDisplay {} {
# Set viewport size to framesize
    GetPlotExtents w h
    PackViewport $w $h 1
#    global canvas ;  $canvas xview moveto 0 ; $canvas yview moveto 0
    SetScrollCenter 0.5 0.5
}

proc FillDisplay {} {
# Set framesize to viewport size
    global FrameScale plot_config
    global ViewportWidth ViewportHeight
    global MinViewportWidth MinViewportHeight

    set w $ViewportWidth
    if { $w < $MinViewportWidth } { set w $MinViewportWidth }
    set h $ViewportHeight
    if { $h < $MinViewportHeight } { set h $MinViewportHeight }

    # Make margins adjustments
    set mgnadj [expr 2*$plot_config(misc,margin)]
    set w  [expr $w-$mgnadj]
    set h  [expr $h-$mgnadj]

    set scale [SetFrameScale $w $h]

    if {$scale!=$FrameScale} {
	global canvas
	OMF_SaveDisplayState
	set FrameScale $scale
	Ow_PushWatchCursor
	DrawFrame $canvas
	PackViewport
	$canvas xview moveto 0
	$canvas yview moveto 0
#	SetScrollCenter 0.5 0.5
	Ow_PopWatchCursor
    }
}

proc ClearFrame {} {
    global ViewportWidth ViewportHeight DisplayRotation
    set autodisplay(sample) 0
    set autodisplay(zoom) 0
    ReadFile {} {} $ViewportWidth $ViewportHeight \
	    $DisplayRotation autodisplay
}

proc ReadFile { filename title {width {}} {height {}} {rot {}} \
	{autodispname {}}} {
    global FileId RealFileId canvas
    global ViewportWidth ViewportHeight DisplayRotation
    global DefaultViewportWidth DefaultViewportHeight
    global FrameScale plot_config

    # Put up watch cursor
    Ow_PushWatchCursor

    # Store current translations
    GetScrollCenter cx cy

    if {[string match {} $width] || $width<3} {
	set width $DefaultViewportWidth
    }
    if {[string match {} $height] || $height<3} {
	set height $DefaultViewportHeight
    }
    if {[string match {} $rot]} {
	set rot $DisplayRotation
    }

    # Set defaults for all display parameters
    if {![string match {} $autodispname]} {
	upvar $autodispname autodisplay
    } else {
	# Default is to enable autodisplay
	set autodisplay(sample) 1
	set autodisplay(zoom)   1
    }

    if {$autodisplay(zoom)} { 
	set scale -1
	set cx 0.5 ; set cy 0.5  ;# Auto-center too.
    } else {
	set scale $FrameScale
    }

    # If auto-sampling is requested, temporarily reconfigure
    # plot_config to invoke auto-sampling.  (Note: The values
    # in plot_config are reset directly into the new mesh, so
    # it is not necessary to call UpdatePlotConfiguration.
    if {$autodisplay(sample)} {
        set arrow_sample_save $plot_config(arrow,subsample)
        set pixel_sample_save $plot_config(pixel,subsample)
	set plot_config(arrow,subsample) -1
	set plot_config(pixel,subsample) -1
    }

    # Use user filter (based on filename extension)?
    set tempfile [OMF_MmdispInputFilter $filename]
    if { [string match {} $tempfile] } {
	# No filter match
	set workname $filename
    } else {
	# Filter match
	set workname $tempfile
    }

    # Read file
    ChangeMesh $workname $width $height $rot $scale
    if {![string match {} $title]} {
	SetMeshTitle $title
    }

    # Delete tempfile, if any
    if { ![string match {} $tempfile] } {
	catch { file delete $tempfile }
    }

    incr FileId
    if {![string match "Vf_EmptyMesh" [GetMeshType]]} {
	incr RealFileId
    }

    if {$ViewportHeight<3} {
	GetPlotExtents ViewportWidth ViewportHeight
    }

    DrawFrame $canvas
    GetFrameRotation DisplayRotation
    GetFrameScale FrameScale
    SetArrowSubsamplingIncrement

    # Subsample values that weren't in auto-sample mode should
    # be removed from auto-sample mode.
    if {$autodisplay(sample)} {
        if {$arrow_sample_save>=0.} {
            set plot_config(arrow,subsample) \
                    [expr abs($plot_config(arrow,subsample))]
            if {$plot_config(arrow,subsample)<=0.01} {
                set plot_config(arrow,subsample) 0
            }
        }
        if {$pixel_sample_save>=0.} {
            set plot_config(pixel,subsample) \
                    [expr abs($plot_config(pixel,subsample))]
            if {$plot_config(pixel,subsample)<=0.01} {
                set plot_config(pixel,subsample) 0
            }
        }
        UpdatePlotConfiguration
    }
    UpdateCoordColors
    PackViewport {} {} 1

    # Automatically fill and wrap first file.
    if {$RealFileId==1} {
	FillDisplay
	WrapDisplay
    }

    # Set to translations to original values.
    SetScrollCenter $cx $cy
    Ow_PopWatchCursor
}

proc SaveFileCallback { widget } {
    global ViewportWidth ViewportHeight
    global MinViewportWidth MinViewportHeight
    global FrameSubsampleIncrement

    set errmsg {}

    # Get filename
    set filename [$widget GetFilename]

    # Store OMF type data & desc strings into omfspec array,
    # using the option name (w/o leading hyphen) as array index.
    foreach { name value } [$widget GetOptionState] {
	set omfspec([string range $name 1 end]) $value
    }
    foreach { name value } [$widget GetDescStrings] {
	set omfspec([string range $name 1 end]) $value
    }

    # Check to see if file already exits
    if { [file exists $filename] } {
	set answer [Ow_Dialog 1 "File Save: File exists" question \
		"File $filename already exists.\nOverwrite?" \
		{} 1 "Yes" "No" ]
	if { $answer != 0 } {
	    lappend errmsg "ERROR: File exists; no overwrite requested."
	    return $errmsg
	}
    }

    # Write file
    Ow_PushWatchCursor
    if { [catch {WriteMesh $filename $omfspec(datatype) $omfspec(gridtype) \
	    $omfspec(title) $omfspec(desc)} writeresult] } {
	lappend errmsg $writeresult
    }
    Ow_PopWatchCursor

    return $errmsg
}

proc OpenFileCallback { widget } {
    global ViewportWidth ViewportHeight
    global MinViewportWidth MinViewportHeight
    global FrameSubsampleIncrement

    set errmsg {}

    # Get filename and "auto" data from File|Open widget
    set filename [$widget GetFilename]
    foreach { name value } [$widget GetOptionState] {
	# Strip leading "-auto" from $name, and stuff
	# value into autodisplay array
	set autodisplay([string range $name 5 end]) $value
    }

    if { ![file exists $filename] } {
        OMF_InfoDialog "File $filename does not exist." \
		"[Oc_Main GetAppName]:Error" warning 10c
	lappend errmsg "ERROR: File does not exists"
    } elseif { ![file isfile $filename] } {
        OMF_InfoDialog "File $filename is not an ordinary file." \
		"[Oc_Main GetAppName]:Error" warning 10c
	lappend errmsg "ERROR: Special file"
    } elseif { ![file readable $filename] } {
        OMF_InfoDialog "Don't have read access to file $filename."
		"[Oc_Main GetAppName]:Error" warning 10c
	lappend errmsg "ERROR: No read access"
    } else {
        GetFrameRotation rot
        # If viewport has 0 area, fall back to default values (as needed).
        global DefaultViewportWidth DefaultViewportHeight
        if {$ViewportWidth<2} {
            set ViewportWidth $DefaultViewportWidth
        }
        if {$ViewportHeight<2} {
            set ViewportHeight $DefaultViewportHeight
        }

        # To be visually consistent, resize ViewportWidth/Height
        # to satisfy lower bounds.
        if {$ViewportWidth<$MinViewportWidth} { 
            set ViewportWidth $MinViewportWidth
        }
        if {$ViewportHeight<$MinViewportHeight} { 
            set ViewportHeight $MinViewportHeight
        }

	# Read file
        ReadFile $filename {} $ViewportWidth $ViewportHeight \
		$rot autodisplay

        # Adjust subsample increment
        set newincr [GetMeshIncrement]
        if {$newincr>0} { set FrameSubsampleIncrement $newincr }
        # Initialize omfSavedDisplayState
        OMF_SaveDisplayState
    }
    return $errmsg
}

proc DialogCallback { widget actionid args } {
    # Re-enable menubutton when the dialog widget is destroyed
    if {[string match DELETE $actionid]} {
        eval [join $args]
	return
    }

    # Call appropriate action proc
    set errmsg "ERROR (proc DialogCallback): Invalid actionid: $actionid"
    if {[string match OPEN $actionid]} {
	set errmsg [OpenFileCallback $widget]
    } elseif {[string match SAVE $actionid]} {
	set errmsg [SaveFileCallback $widget]
    }

    return $errmsg
}

array set displayprn {
    orient   "landscape"
    paper    "letter"
    hpos     "center"
    vpos     "center"
    units    "in"
    tmargin  "1.0"
    lmargin  "1.0"
    pwidth   "6.0"
    pheight  "6.0"
    crop     "1"
    aspect_crop "1"
    aspect_nocrop "1"
}

proc UpdatePrintAspectRatios { args } {
    global displayprn
    # Set non-cropped aspect ratio
    GetPlotExtents width height
    if { $height>0 && $width>0 } {
	set aspect [expr double($width)/double($height)]
	if {$displayprn(aspect_nocrop)!=$aspect} {
	    set displayprn(aspect_nocrop) $aspect
	}
    }
    # Set cropped aspect ratio
    global CanvasWidth CanvasHeight
    if { $CanvasHeight>0 && $CanvasWidth>0 } {
	set aspect [expr double($CanvasWidth)/double($CanvasHeight)]
	if {$displayprn(aspect_crop)!=$aspect} {
	    set displayprn(aspect_crop) $aspect
	}
    }
}

trace variable UpdateViewCount w UpdatePrintAspectRatios

proc PrintCallback { printwidget arrname } {
    upvar $arrname dialogprn
    global canvas displayprn
    if {[info exists dialogprn]} {
	array set displayprn [array get dialogprn]
    }
    Ow_PushWatchCursor
    set errcode [catch \
	    {Ow_PrintDlg Print $printwidget $canvas displayprn} msg]
    Ow_PopWatchCursor
    if {$errcode} {
	Oc_Log Log $msg error
    }
}

proc LaunchDialog { menuname itemlabel action } {
    if { [string match [$menuname entrycget $itemlabel -state] "disabled"] } {
	return   ;# Item not enabled
    }
    global FileId

    if {[string match OPEN $action]} {
	OpenDlg New dialogbox -callback DialogCallback \
		-dialog_title "[Oc_Main GetAppName]:Open File" \
		-filter "*.o?f *.svf *.vio" \
		-delete_callback_arg \
		[list $menuname entryconfigure $itemlabel -state normal]
	$menuname entryconfigure $itemlabel -state disabled
    } elseif {[string match SAVE $action]} {
	SaveDlg New dialogbox -callback DialogCallback \
		-dialog_title "[Oc_Main GetAppName]:Save File" \
		-filter "*.o?f *.svf *.vio" \
		-mesh_update_variable FileId \
		-mesh_update_proc OMF_MeshInfo \
		-delete_callback_arg \
		[list $menuname entryconfigure $itemlabel -state normal]
	$menuname entryconfigure $itemlabel -state disabled
    } elseif {[string match PRINT $action]} {
        global displayprn
	Ow_PrintDlg New dialogbox \
                -apply_callback PrintCallback \
		-dialog_title "[Oc_Main GetAppName]:Print" \
                -menu [list $menuname $itemlabel] \
                -crop_option 1 \
                -import_arrname displayprn
    } else {
	Ow_NonFatalError "Invalid action code: $action" \
		"Error in LaunchDialog proc"
	return
    }

    # Set icon
    Ow_SetIcon [$dialogbox Cget -winpath]
}

proc UpdateCoordColors { args } {
    global plot_config coords_win
    set bgcolor $plot_config(misc,background)
    set axiscolors [list \
            [GetVecColor [list  1.  0.  0.]] \
            [GetVecColor [list -1.  0.  0.]] \
            [GetVecColor [list  0.  1.  0.]] \
            [GetVecColor [list  0. -1.  0.]] \
            [GetVecColor [list  0.  0.  1.]] \
            [GetVecColor [list  0.  0. -1.]]]
    SetCoordColors $coords_win $bgcolor $axiscolors    
}

proc SetPlotConfiguration { cfgname } {
    global plot_config canvas
    upvar $cfgname cfg
    # Take appropriate actions
    array set plot_config [array get cfg]
    Ow_PushWatchCursor
    UpdatePlotConfiguration
    DrawFrame $canvas
    UpdateCoordColors
    PackViewport
    Ow_PopWatchCursor
}

###################### WIDGETS (Definitions & Bindings) #############
# Canvas widget
set canvas_frame [frame .canvas_frame -borderwidth 0]
set canvas [canvas $canvas_frame.canvas -background white -confine 0 \
        -borderwidth 0 -highlightthickness 0 -selectborderwidth 0 \
        -xscrollcommand "$canvas_frame.xscroll set" \
        -yscrollcommand "$canvas_frame.yscroll set" \
	-scrollregion "0 0 1 1" \
        -width 0 -height 0]
# Note: -confine is turned off so translations can take place without
#  requiring an update of the scrollbars and scrollregion (via a call
#  to 'update idletasks').  In particular, this allows for cleaner
#  rotation processing.
#    The following scroll filter restores -confine type behavior for
#  user scrollbar interaction.
proc CanvasConfine { args } {
    global canvas plot_config
    if {$plot_config(misc,confine)} {
	$canvas configure -confine 1
	eval $canvas $args
	$canvas configure -confine 0
    } else {
	eval $canvas $args
    }
}

frame $canvas_frame.yscrollbox -borderwidth 0 ;# This needs to be created
                       # *before* yscroll so it gets mapped under yscroll.
scrollbar $canvas_frame.xscroll -troughcolor $omfTroughColor \
           -command "CanvasConfine xview"  -orient horizontal
scrollbar $canvas_frame.yscroll -troughcolor $omfTroughColor \
           -command "CanvasConfine yview"  -orient vertical
pack $canvas -side left -fill both -expand 1
set ScrollCrossDim [expr \
        [$canvas_frame.yscroll cget -width] \
        + 2*[$canvas_frame.yscroll cget -borderwidth] \
        + 2*[$canvas_frame.yscroll cget -highlightthickness]]
frame $canvas_frame.yscrollbox.corner -borderwidth 2 -relief raised \
        -width $ScrollCrossDim -height $ScrollCrossDim
pack $canvas_frame.yscrollbox.corner -side bottom
# The yscrollbox.corner is used if both x & y scrollbars are displayed.
# By making the scrollbars symmetric, it eases the calculations needed
# to keep the same part of the image displayed under rotations.
# NOTE: The yscroll widget is directly packed into $canvas_frame if
#    xscroll is not displayed, but packed into $canvas_frame.yscrollbox
#    if xscroll is displayed.  See 'proc UpdateScrollBarDisplay'.

# NOTE: $canvas <Configure> binding is set below in
#       INITIALIZATION section, and may get temporarily unset
#       and set in other places as well.

# Option frame and widgets
set ctrlbar [frame .ctrlbar -borderwidth 2 -relief raised]
trace variable omfCtrlBarState w OMF_ChangeCtrlBarState
proc OMF_ChangeCtrlBarState { name elt op } {
    global omfCtrlBarState ctrlbar canvas_frame
    if {$omfCtrlBarState} {
        pack $ctrlbar -side top -before $canvas_frame -fill x -expand 0
    } else {
        pack forget $ctrlbar
	focus $canvas_frame     ;# Shift focus off $ctrlbar
    }
    OMF_SetGeometryRequirements
}

# Rotation
frame $ctrlbar.rotation -borderwidth 2 -relief ridge
set coords_win [canvas $ctrlbar.rotation.coords \
        -width 48 -height 0 -bg white]
# The -width 48 is just a size guess; the canvas gets enlarged
# to a square region that just fills the height required
# by the other control bar lines. NOTE: The -height 0 value
# seems likely to be best behaved value wrt the <Configure>
# bindings.
pack $ctrlbar.rotation.coords -side left -fill y
proc UpdateCoordDisplay { newrot } {
    global ctrlbar
    set win $ctrlbar.rotation.coords
    ShowCoords $win $newrot
}
trace variable DisplayRotation w DisplayRotationTrace
bind $ctrlbar.rotation.coords <Button-1> {
    set temp [expr (round($DisplayRotation/90.)+1)%%4]
    set DisplayRotation [expr $temp*90]
}
bind $ctrlbar.rotation.coords <Key-Left> {
    # Same as <Button-1>
    set temp [expr (round($DisplayRotation/90.)+1)%%4]
    set DisplayRotation [expr $temp*90]
}
bind $ctrlbar.rotation.coords <Button-3> {
    set temp [expr (round($DisplayRotation/90.)-1)%%4]
    set DisplayRotation [expr $temp*90]
}
bind $ctrlbar.rotation.coords <Key-Right> {
    # Same as <Button-3>
    set temp [expr (round($DisplayRotation/90.)-1)%%4]
    set DisplayRotation [expr $temp*90]
}
bind $ctrlbar.rotation.coords <Configure> {
    set bw [%W cget -borderwidth]
    set ht [%W cget -highlightthickness]
    set size [expr %h - 2*$bw - 2*$ht]
    set cwidth [%W cget -width]
    set cheight [%W cget -height]
    if {$cwidth!=$size || $cheight!=$size} {
        %W configure -width $size -height $size
    }
}

# Control bar row 1 frame; Holds "Subsample" and "Arrow mag" controls
frame $ctrlbar.row1

# Arrow subsampling
proc SetArrowSubsampling { subsample } {
    global plot_config canvas
    if { $subsample == $plot_config(arrow,subsample) } { return }
    set plot_config(arrow,subsample) $subsample
    Ow_PushWatchCursor
    UpdatePlotConfiguration
    DrawFrame $canvas
    Ow_PopWatchCursor
}

Ow_EntryScale New arrowsswidget $ctrlbar.subsample -label "Arrow Subsample:" \
        -variable plot_config(arrow,subsample) -valuewidth 5 \
        -scalewidth 50 -scalemin 0. -scalemax 20. -scalestep 0.5 \
	-troughcolor $_mmdisp(troughcolor) \
        -outer_frame_options "-bd 2 -relief ridge" \
	-command SetArrowSubsampling

proc SetArrowSubsamplingIncrement { {incr {}} } {
    global arrowsswidget
    if {[string match {} $incr]} { set incr [GetMeshIncrement] }
    $arrowsswidget Configure -scalestep $incr
}

# Arrow mag
proc SetArrowMag { arrowmag } {
    global plot_config canvas
    if { $arrowmag == $plot_config(arrow,mag) } { return }
    set plot_config(arrow,mag) $arrowmag
    Ow_PushWatchCursor
    UpdatePlotConfiguration
    DrawFrame $canvas
    Ow_PopWatchCursor
}

proc ArrowMagCallback { widget action args } {
    global plot_config
    set x [$widget Cget -value]
    set y [OMF_MakeValidNumber $x]
    if { $x != $y } {
        # Illegal input string
        $widget Set $plot_config(arrow,mag)
    } else {
        SetArrowMag $x
    }
}

Ow_EntryBox New arrowmagwidget $ctrlbar -label "Mag:" \
        -callback ArrowMagCallback -autoundo 0 \
        -valuewidth 5 -variable plot_config(arrow,mag) \
	-valuetype fixed -coloredits 1 -writethrough 0 \
        -outer_frame_options "-bd 2 -relief ridge"

# Frame scale
Ow_EntryScale New zoomwidget $ctrlbar.framescale -label "Zoom:" \
        -variable FrameScale -valuewidth 5 \
        -scalewidth 50 -scalemin 0. -scalemax 100. -scalestep .5 \
        -outer_frame_options "-bd 2 -relief ridge" \
	-troughcolor $_mmdisp(troughcolor) \
        -command SetScale

### $ctrlbar packing
pack $ctrlbar.rotation   -side left -fill y
pack $ctrlbar.row1       -side top -fill x
pack [$arrowsswidget Cget -winpath]  -in $ctrlbar.row1 \
	-side left  -fill both -expand 1 
pack [$arrowmagwidget Cget -winpath] -in $ctrlbar.row1 -side right -fill y
pack [$zoomwidget Cget -winpath] -side top -fill x

########################### MENUS ###################################
set menubar .menu
foreach {filemenu viewmenu optionmenu helpmenu} \
	[Ow_MakeMenubar . $menubar File View Options Help] {}

$filemenu add command -label "Open..." -underline 0 \
        -command { LaunchDialog $filemenu "Open..." OPEN }
## Note: If you change this menu item, then you must also change the
##       key binding with Ctrl-o; See the section "ADDITIONAL KEY BINDINGS"
$filemenu add command -label "Clear" -underline 0 -command { ClearFrame }
$filemenu add command -label "Save as..." -underline 0 \
	-command { LaunchDialog $filemenu "Save as..." SAVE }
## Note: If you change this menu item, then you must also change the
##       key binding with Ctrl-s; See the section "ADDITIONAL KEY BINDINGS"
$filemenu add command -label "Print..." -underline 0 \
        -command { LaunchDialog $filemenu "Print..." PRINT }
## Note: If you change this menu item, then you must also change the
##       key binding with Ctrl-p; See the section "ADDITIONAL KEY BINDINGS"
$filemenu add separator
$filemenu add command -label "Exit" -command { exit } -underline 1
wm protocol . WM_DELETE_WINDOW { exit }

$viewmenu add command -label "Wrap Display" -command { WrapDisplay } \
        -underline 0
$viewmenu add command -label "Fill Display" -command { FillDisplay } \
        -underline 0
$viewmenu add command -label "Rotate ccw" -underline 0 \
        -command {set DisplayRotation [expr round($DisplayRotation+90)%360]}
$viewmenu add command -label "Rotate cw" -underline 7 \
        -command {set DisplayRotation [expr round($DisplayRotation-90)%360]}
$viewmenu add command -label "reDraw" -underline 2 -command { RedrawDisplay }

$optionmenu add command -label "Configure..." -underline 0 \
        -command { Mmd_PlotSelectConfigure New _ \
                  -menu "$optionmenu {Configure...}" \
                  -import_arrname plot_config \
                  -apply_callback SetPlotConfiguration
              }
$optionmenu add checkbutton -label "Control Bar" -underline 0 \
        -variable omfCtrlBarState

Ow_StdHelpMenu $helpmenu

if {$tk_version>=8.0 && [string match windows $tcl_platform(platform)]} {
    # Windows doesn't size Tcl 8.0 menubar cleanly
    wm geometry . "[expr [Ow_GuessMenuWidth $menubar]+25]x0"
}


####################### ADDITIONAL KEY BINDINGS #####################
# Allow the canvas frame to acquire keyboard focus, and prohibit
# canvas or canvas scrollbars from same (all necessary key bindings
# are bound to the canvas frame.
set cf $canvas_frame
$cf.xscroll configure -takefocus 0
$cf.yscroll configure -takefocus 0
$cf.canvas  configure -takefocus 0
$cf configure -takefocus 1

proc Zoom { amt } { global FrameScale; SetScale [expr $amt*$FrameScale] }
bind $cf <Key-Escape>      { OMF_RecoverDisplayState }
bind $cf <Key-Prior>       { Zoom $SmallZoom }
bind $cf <Shift-Key-Prior> { Zoom $LargeZoom }
bind $cf <Key-Next>        { Zoom [expr 1./$SmallZoom] }
bind $cf <Shift-Key-Next>  { Zoom [expr 1./$LargeZoom] }
## NOTE: SunOS apparently doesn't have Page_Up & Page_Down.  For the
## record, Prior == Page_Up, and Next == Page_Down.
# Ditto for KeyPad; Note that the <Shift> keypad values are funky,
# and might be wrong for some keyboards.
catch {bind $cf <Key-KP_Prior>    { Zoom $SmallZoom }}
catch {bind $cf <Shift-Key-KP_9>  { Zoom $LargeZoom }}
catch {bind $cf <Key-KP_Next>     { Zoom [expr 1./$SmallZoom] }}
catch {bind $cf <Shift-Key-KP_3>  { Zoom [expr 1./$LargeZoom] }}

bind $cf <Control-Key-f> { FillDisplay }
bind $cf <Control-Key-w> { WrapDisplay }
bind $cf <Key-Home>      { FillDisplay ; WrapDisplay }
catch {bind $cf <Key-KP_Home>   { FillDisplay ; WrapDisplay }}
bind $cf <Control-Key-d> { RedrawDisplay }
bind $cf <Control-Key-o> { LaunchDialog $filemenu "Open..." OPEN }
bind $cf <Control-Key-s> { LaunchDialog $filemenu "Save as..." SAVE }
bind $cf <Control-Key-p> { LaunchDialog $filemenu "Print..." PRINT }
## Note: The last two depend upon the text in the pull down menus!

bind $cf <Control-Key-r> { ;# Ctrl-r
    set DisplayRotation [expr round($DisplayRotation+90)%%360]
}
bind $cf <Control-Key-R> { ;# Shift-Ctrl-R
    set DisplayRotation [expr round($DisplayRotation-90)%%360]
}
bind $cf <Control-Key-c> { ;# Ctrl-c
    set DisplayRotation [expr round($DisplayRotation-90)%%360]
}
bind $cf <Key-Insert> { 
    set newrate [expr $plot_config(arrow,subsample)-1]
    if {$newrate<0} {set newrate 0}
    SetArrowSubsampling $newrate
}
bind $cf <Key-Delete> {
    SetArrowSubsampling [expr $plot_config(arrow,subsample)+1]
}
bind $cf <Shift-Key-Insert> {
    SetArrowSubsampling [expr $plot_config(arrow,subsample)/2.]
}
bind $cf <Shift-Key-Delete> {
    SetArrowSubsampling [expr $plot_config(arrow,subsample)*2]
}
# Ditto for KeyPad; Note that the <Shift> keypad values are funky,
# and might be wrong for some keyboards.
catch { bind $cf <Key-KP_Insert> { 
    set newrate [expr $plot_config(arrow,subsample)-1]
    if {$newrate<0} {set newrate 0}
    SetArrowSubsampling $newrate
}   }
catch {bind $cf <Key-KP_Delete> {
    SetArrowSubsampling [expr $plot_config(arrow,subsample)+1]
}   }
catch {bind $cf <Shift-Key-KP_0> {
    SetArrowSubsampling [expr $plot_config(arrow,subsample)/2.]
}   }
catch {bind $cf <Shift-Key-KP_Decimal> {
    SetArrowSubsampling [expr $plot_config(arrow,subsample)*2]
}   }

# Proc to scroll canvas from keyboard.  direction should be one of
# left|right|up|down, and stride should be one of units|pages.
proc KeyboardScrollCanvas { direction stride } {
    global canvas
    switch -- $direction {
        left     { CanvasConfine xview scroll -1 $stride }
        right    { CanvasConfine xview scroll  1 $stride }
        up       { CanvasConfine yview scroll -1 $stride }
        down     { CanvasConfine yview scroll  1 $stride } 
        default  {
            Ow_FatalError "Invalid direction ($direction) detected in \
                    proc KeyboardScrollCanvas"
        }
    }
}
bind $cf <Key-Left>  { KeyboardScrollCanvas left  units }
bind $cf <Key-Right> { KeyboardScrollCanvas right units }
bind $cf <Key-Up>    { KeyboardScrollCanvas up    units }
bind $cf <Key-Down>  { KeyboardScrollCanvas down  units }
bind $cf <Shift-Key-Left>  { KeyboardScrollCanvas left  pages }
bind $cf <Shift-Key-Right> { KeyboardScrollCanvas right pages }
bind $cf <Shift-Key-Up>    { KeyboardScrollCanvas up    pages }
bind $cf <Shift-Key-Down>  { KeyboardScrollCanvas down  pages }
# Ditto for KeyPad; Note that the <Shift> keypad values are funky,
# and might be wrong for some keyboards.
catch {bind $cf <Key-KP_Left>  { KeyboardScrollCanvas left  units }}
catch {bind $cf <Key-KP_Right> { KeyboardScrollCanvas right units }}
catch {bind $cf <Key-KP_Up>    { KeyboardScrollCanvas up    units }}
catch {bind $cf <Key-KP_Down>  { KeyboardScrollCanvas down  units }}
catch {bind $cf <Shift-Key-KP_4> { KeyboardScrollCanvas left  pages }}
catch {bind $cf <Shift-Key-KP_6> { KeyboardScrollCanvas right pages }}
catch {bind $cf <Shift-Key-KP_8> { KeyboardScrollCanvas up    pages }}
catch {bind $cf <Shift-Key-KP_2> { KeyboardScrollCanvas down  pages }}

# Try to handle <Shift-Key-Tab> variants.  The [bind all <Shift-Tab>]
# parameter expands to the proc that is the current binding to
# <Shift-Tab> (probably 'tkTabToWindow [tk_focusPrev %W]').  Some
# X servers return a distinct keysym for the Shift-Tab event, e.g.,
# ISO_Left_Tab _instead_ Tab.  One can use the X utility program
# xev to see what keysyms are being generated.
#  One problem that occurs, is if one is running a binary on one
# machine, say, a Sun, that does not know the ISO_Left_Tab keysym
# (cf. the system <X11/keysymdef.h> file), and displaying on a
# different system, say recent Linux machine, that does.  Then the
# X server on the Linux machine generates the keysym number corresponding
# to ISO_LEFT_TAB (0xfe20), but the program built on the Sun doesn't
# know the symbolic interpretation for that keysym.  Therefore we
# provide a numeric binding if the symbolic binding is not known.  (I'm
# assuming the symbolic bindings, if defined, are the same across all
# platforms.)
#  In addition to xev, the following wish script is useful in diagnosing
# and solving such problems:
#
# bind . <KeyPress> {puts stdout {%%K=%K, %%A=%A, %%k=%k, %%s=%s, %%N=%N.}}
#
# This prints to stdout the  keysym (%K), ASCII character (%A), keycode (%k)
# state (%s, which for key events is modifier information, like shift key
# status), and keysym as a decimal number (%N).  The last should be defined
# even when the first isn't.          -mjd, Sep-1997
#
if {[catch {bind all <Shift-Key-ISO_Left_Tab> [bind all <Shift-Tab>]}]} {
    # ISO_Left_Tab probably not defined.  Try numeric binding
    catch {
	bind all <Shift-Any-Key> \
		"+if {\"%N\"==65056} \{[bind all <Shift-Tab>]\}"
	# ISO_Left_Tab = 0xFE20 = 65056
	# Man, talk about quoting hell.  We need the braces on the
	# second term of the "if" to keep "if" happy, but we need
	# to get the [bind all <Shift-Tab>] expanded at bind time to
	# get the imbedded %W (from inside [bind all <Shift-Tab]) in
	# place for binding substitution...
    }
}
if {[catch {bind all <Shift-Key-3270_BackTab> [bind all <Shift-Tab>]}]} {
    # 3270_BackTab probably not defined.  Try numeric binding
    catch {
	bind all <Shift-Any-Key> \
		"+if {\"%N\"==64773} \{[bind all <Shift-Tab>]\}"
	# 3270_BackTab = 0xFD05 = 64773
	# See the notes in the previous stanza about "quoting hell."
    }
}

###################### ADDITIONAL MOUSE BINDINGS ####################
### Canvas rectangle zoom-in
# Zoom box foreground line color, A -> enlarge, B -> contract
set omfZoomBoxForeColorA [Ow_GetRGB red]
set omfZoomBoxForeColorB [Ow_GetRGB MediumBlue]
# For reasons unknown to me (mjd 97/2/22), when drawing on an 8-bit
# pseudocolor visual with a shared colormap, under interactive
# conditions (redrawing zoom boxes with mouse motion) it took noticeable
# longer to erase and draw "red" boxes than "MediumBlue" boxes.  I
# discovered that the difference goes away if "red" is changed to, e.g.,
# "IndianRed", or to the form #xxx.  I also discovered that you can
# apparently find out what color you will actually get with the command
# "winfo rgb . color", where color can be in any of the usual Tk color
# formats.  For example,
#   %winfo rgb . MediumBlue
#  > 0 0 52685
#   %winfo rgb . #FF0000
#  > 65535 0 0
#   %winfo rgb . #FE0000
#  > 65535 0 0
#
set omfZoomBoxForeThick  1       ;# Zoom box foreground line thickness
set omfZoomBoxBackColor  white   ;# Zoom box background line color
set omfZoomBoxBackThick  3       ;# Zoom box background line thickness
set omfZoomBoxMinWidth   5;# This must be >0.
set omfZoomBoxArrowSize  12
set omfZoomDirection     0       ;# When active, should be +1 => enlarge
                                  #                     or -1 => contract
set omfZoomBoxLoiterTime 3000    ;# Number of milliseconds to keep zoom box
                                  # visible after zooming.
proc OMF_InitZoomBox { v1x v1y direction } {
    global canvas omfP1x omfP1y omfZoomDirection
    if {$omfZoomDirection!=0} return  ;# Zoom (other direction) already
    set omfZoomDirection $direction   ;# in progress
    set omfP1x [$canvas canvasx $v1x]
    set omfP1y [$canvas canvasy $v1y]
}

proc OMF_StrokeZoomBox { x1 y1 x2 y2 boxtag} {
    global omfZoomBoxBackColor omfZoomBoxBackThick
    global omfZoomBoxForeThick omfZoomDirection
    global omfZoomBoxForeColorA omfZoomBoxForeColorB
    global canvas
    if {$omfZoomDirection==0} return
    if {$omfZoomDirection>0} { set forecolor $omfZoomBoxForeColorA } \
            else             { set forecolor $omfZoomBoxForeColorB }
    $canvas create rectangle $x1 $y1 $x2 $y2 \
            -outline $omfZoomBoxBackColor -width $omfZoomBoxBackThick \
            -tag $boxtag
    $canvas create rectangle $x1 $y1 $x2 $y2 \
            -outline $forecolor -width $omfZoomBoxForeThick \
            -tag $boxtag
}

proc OMF_DrawZoomBox { v2x v2y } {
    global FrameScale zoomwidget
    global omfP1x omfP1y canvas
    global omfZoomBoxMinWidth
    global omfZoomDirection omfZoomBoxForeColorA omfZoomBoxForeColorB
    if {$omfZoomDirection==0} return
    if {$omfZoomDirection>0} {
        set forecolor $omfZoomBoxForeColorA
    } else {
        set forecolor $omfZoomBoxForeColorB
    }
    global omfZoomBoxBackColor omfZoomBoxForeThick omfZoomBoxBackThick
    global omfZoomBoxArrowSize
    $canvas delete zoombox ;# erase old zoom box
    set p2x [$canvas canvasx $v2x]
    set p2y [$canvas canvasy $v2y]
    set zw [expr abs($p2x-$omfP1x)]
    set zh [expr abs($p2y-$omfP1y)]
    if {$zw<$omfZoomBoxMinWidth || $zh<$omfZoomBoxMinWidth} {
	$zoomwidget Set $FrameScale
        return
    }
    OMF_StrokeZoomBox $omfP1x $omfP1y $p2x $p2y zoombox

    # Determine which side (x- or y-) will limit expansion
    global CanvasWidth CanvasHeight
    set vw $CanvasWidth
    set vh $CanvasHeight
    if {$omfZoomDirection>0} {
        # Make adjustments to viewport dimensions under the
        # assumption that any non-active scroll bars will be
        # turn on as a result of the zoom.
        global XScrollState YScrollState ScrollCrossDim
        if {!$XScrollState} {
            set vh [expr $vh-$ScrollCrossDim]
        }
        if {!$YScrollState} {
            set vw [expr $vw-$ScrollCrossDim]
        }
    }
    if {$vw<1} {set vw 1} ; if {$vh<1} {set vh 1} ;# Safety
    set xlimit [expr $vw*$zh]; set xfudge [expr 2*$vw]
    set ylimit [expr $vh*$zw]; set yfudge [expr 2*$vh]
    if {$omfZoomDirection<0} {
        set temp $xlimit; set xlimit $ylimit; set ylimit $temp;
        set temp $xfudge; set xfudge $yfudge; set yfudge $temp;
    }

    # Draw zoom direction arrows on expansion limit sides
    if {$omfP1x<$p2x} { set xmin $omfP1x; set xmax $p2x    } \
            else      { set xmin $p2x;    set xmax $omfP1x }
    if {$omfP1y<$p2y} { set ymin $omfP1y; set ymax $p2y    } \
            else      { set ymin $p2y;    set ymax $omfP1y }
    set xmid [expr round(($xmin+$xmax)/2.)]
    set ymid [expr round(($ymin+$ymax)/2.)]
    set arrlength [expr round($omfZoomBoxArrowSize)]
    if {$omfZoomDirection>=0} { set offset $arrlength } \
            else              { set offset -$arrlength }
    set arrheadsize [expr round($arrlength/2)]
    set arrheadshape "$arrheadsize  $arrheadsize [expr $arrheadsize/2]"
    if { $ylimit < $xlimit + $xfudge} {
        # Draw arrows on top & bottom sides
        $canvas create line \
                $xmid [expr $ymin+$offset] $xmid $ymin \
                -fill $forecolor -width $omfZoomBoxForeThick \
                -arrow last -arrowshape $arrheadshape -tag zoombox
        $canvas create line \
                $xmid [expr $ymax-$offset] $xmid $ymax \
                -fill $forecolor -width $omfZoomBoxForeThick \
                -arrow last -arrowshape $arrheadshape -tag zoombox
    }
    if { $xlimit < $ylimit + $yfudge} {
        # Draw arrows on left & right sides
        $canvas create line \
                [expr $xmin+$offset] $ymid $xmin $ymid \
                -fill $forecolor -width $omfZoomBoxForeThick \
                -arrow last -arrowshape $arrheadshape -tag zoombox
        $canvas create line \
                [expr $xmax-$offset] $ymid $xmax $ymid \
                -fill $forecolor -width $omfZoomBoxForeThick \
                -arrow last -arrowshape $arrheadshape -tag zoombox
    }
    # Reflect anticipated zoom state in scale widgets
    if {$vw>0 && $vh>0 && $zw>0 && $zh>0} { ;# Safety check
        if {$ylimit < $xlimit} {
            set scale [expr double($vh)/double($zh)]
        } else {
            set scale [expr double($vw)/double($zw)]
        }
        if {$omfZoomDirection<0} { set scale [expr 1./double($scale)] }
        global omfStdNumericFormat
	set newsize [OMF_FormatNumber [expr $FrameScale*$scale]]
	$zoomwidget SetEdit $newsize
    }
}
proc OMF_DoZoom { v2x v2y } {
    global canvas omfP1x omfP1y
    global FrameScale
    global omfZoomBoxMinWidth omfZoomDirection
    global CanvasWidth CanvasHeight
    global XScrollState YScrollState ScrollCrossDim
    if {$omfZoomDirection==0} return

    set clean_return "$canvas delete zoombox ; return"

    set p2x [$canvas canvasx $v2x]
    set p2y [$canvas canvasy $v2y]
    set zwidth  [expr abs($p2x-$omfP1x)]
    set zheight [expr abs($p2y-$omfP1y)]
    if {$zwidth<$omfZoomBoxMinWidth \
            || $zheight<$omfZoomBoxMinWidth} {
        unset omfP1x omfP1y
        set omfZoomDirection 0
        eval $clean_return
    }
    set vwidth  $CanvasWidth
    set vheight $CanvasHeight
    set xscrollstate_orig $XScrollState
    set yscrollstate_orig $YScrollState
    if {$omfZoomDirection>0} {
        # Make adjustments to viewport dimensions under the
        # assumption that any non-active scroll bars will be
        # turned on as a result of the zoom.
        if {!$XScrollState} {
            set vheight [expr $vheight-$ScrollCrossDim]
        }
        if {!$YScrollState} {
            set vwidth [expr $vwidth-$ScrollCrossDim]
        }
    }
    set xscale [expr double($vwidth)/double($zwidth)]
    set yscale [expr double($vheight)/double($zheight)]
    if {$xscale<0.001 || $xscale>1000  || $yscale<0.001 || $yscale>1000} {
        # Safety check
        unset omfP1x omfP1y
        set omfZoomDirection 0
        eval $clean_return
    }
    if {$omfZoomDirection<0} {
        set xscale [expr 1./double($xscale)]
        set yscale [expr 1./double($yscale)]
    }
    if {$xscale<$yscale} {
        set scale $xscale
    } else {
        set scale $yscale
    }
    # Note: $scale is magnification relative to current $FrameScale;
    #       $FrameScale is magnification relative to mesh scaling.

    OMF_SaveDisplayState
    set newframescale [expr floor($scale*$FrameScale*100.)/100.]
    if {$newframescale>0.} {
	set FrameScale $newframescale
	SetFrameScale $FrameScale
    } else {
        unset omfP1x omfP1y
        set omfZoomDirection 0
        eval $clean_return
    }

    Ow_PushWatchCursor
    $canvas delete zoombox ;# Put old zoom box erase command on event
    ## loop *after* 'update idletask' call inside Ow_PushWatchCursor.
    DrawFrame $canvas
    PackViewport

    # Center middle of zoom box.  If a scrollbar has been turned
    # on or off, the canvas widget won't know about it until after
    # an update idletask call.  To remove the display+scroll action,
    # handle this case by adjusting the centering request.
    set a1x [expr $omfP1x*$scale]
    set a1y [expr $omfP1y*$scale]
    set a2x [expr $p2x*$scale]
    set a2y [expr $p2y*$scale]
    set acenterx [expr ($a1x+$a2x)/2.]
    set acentery [expr ($a1y+$a2y)/2.]
    set xscrolladj [expr ($XScrollState-$xscrollstate_orig)*$ScrollCrossDim/2.]
    set yscrolladj [expr ($YScrollState-$yscrollstate_orig)*$ScrollCrossDim/2.]
    CenterPlotView [expr $acenterx+$xscrolladj] [expr $acentery+$yscrolladj]

    # Draw zoom "loiter" box
    if {$omfZoomDirection>0} {
        # Adjust coordinates so box fits inside viewport
        if {$a2x<$a1x} {set temp $a1x; set a1x $a2x; set a2x $temp}
        if {$a2y<$a1y} {set temp $a1y; set a1y $a2y; set a2y $temp}
	foreach {xmin ymin xmax ymax} [GetCanvasBox] {}
	set xmax [expr $xmax - 1.0]
	set ymax [expr $ymax - 1.0]
        if {$a1x<$xmin} {set a1x $xmin}
        if {$a1y<$ymin} {set a1y $ymin}
        if {$a2x>$xmax} {set a2x $xmax}
        if {$a2y>$ymax} {set a2y $ymax}
        set a1x [expr  ceil($a1x)] ;   set a1y [expr  ceil($a1y)]
        set a2x [expr floor($a2x)] ;   set a2y [expr floor($a2y)]
    }
    global UpdateViewCount
    set boxtag "afterzoombox$UpdateViewCount"
    # The above creates an essentially unique tagname for the zoom
    # loiter box.  If all loiter boxes shared the same tag, then if
    # the user did a second zoom too quick, then the second zoom loiter
    # box could be deleted by the command deleting the first.  OK, this
    # is not a serious problem, but its easy to fix, and I'm here now
    # anyway...
    OMF_StrokeZoomBox $a1x $a1y $a2x $a2y $boxtag

    # Clean up
    unset omfP1x omfP1y
    set omfZoomDirection 0
    global omfZoomBoxLoiterTime
    after $omfZoomBoxLoiterTime "$canvas delete $boxtag"
    Ow_PopWatchCursor
}
proc CenterPlotView { {px {}} {py {}}} {
    foreach {xmin ymin xmax ymax} [GetPlotBox] {}
    if { $xmax<=$xmin || $ymax<=$ymin } { return }
    if {[string match {} $px]} { set px [expr ($xmin+$xmax)/2.] }
    if {[string match {} $py]} { set py [expr ($ymin+$ymax)/2.] }
    set xcenter [expr ($px-$xmin)/($xmax-$xmin)]
    if {$xcenter<0} {set xcenter 0}
    if {$xcenter>1} {set xcenter 1}
    set ycenter [expr ($py-$ymin)/($ymax-$ymin)]
    if {$ycenter<0} {set ycenter 0}
    if {$ycenter>1} {set ycenter 1}
    SetScrollCenter $xcenter $ycenter
}
proc CenterScreenView { vx vy } {
    global canvas
    CenterPlotView [$canvas canvasx $vx] [$canvas canvasy $vy]
}
bind $canvas <Button-1> { OMF_InitZoomBox %x %y 1 }
bind $canvas <Button-3> { OMF_InitZoomBox %x %y -1 }
bind $canvas <Button1-Motion> [Oc_SkipWrap {OMF_DrawZoomBox %x %y}]
bind $canvas <Button3-Motion> [Oc_SkipWrap {OMF_DrawZoomBox %x %y}]
bind $canvas <ButtonRelease> { OMF_DoZoom %x %y }
bind $canvas <ButtonRelease-2> { CenterScreenView %x %y}
bind $canvas <Shift-ButtonRelease-1> { CenterScreenView %x %y}
# Hmm, can't seem to find a home for a CenterView binding

########################### TOP_LEVEL PACKING #######################
pack $canvas_frame -side top -fill both -expand 1
pack propagate $canvas_frame 0
# Note 1: Any extra vertical spacing is divvied up equally among
#   all the first level widgets that have the -expand 1 option.
#   The net result is that frame-canvas resizing won't work properly
#   if any widget besides $canvas_frame has -expand 1.
# Note 2: The outer frame is used as a geometry size support for all
#   its children.  Therefore it is important that propagation be
#   turned off.

# Ctrlbar
if {$omfCtrlBarState} {
    pack $ctrlbar -side top -fill x -expand 0 -before $canvas_frame
}

########################### INITIALIZATION ##########################
# Determine minimum viewport size, and get fixed widget geometry values
OMF_SetGeometryRequirements

# Initialize global variables
set DefaultViewportWidth  [expr $DefaultWindowWidth-10] 
set DefaultViewportHeight [expr $DefaultWindowHeight-$RootHeadHeight-20]
# The '-10' & '-20' in the above lines is just so we don't fill up the
# *whole* display screen (in case the display *is* only 640x480).

# Bind canvas resize proc
bind $canvas_frame <Configure> {
    set border [expr 2*[%W cget -borderwidth]+2*[%W cget -highlightthickness]]
    ResizeViewport [expr %w-$border] [expr %h-$border]
    # Note: The <Configure> resize request *includes* both -borderwidth
    #       & -highlightthickness!
}

# Extract working directory and filename from command line argument
set firstpath {}
set firstfile [GetFirstFileName]
if [file isdirectory $firstfile] {
    # No file, just directory
    set firstpath $firstfile
    set firstfile {}
} else {
    # File and potentially directory
    set firstpath [file dirname $firstfile]
    set firstfile [file tail $firstfile]
}

# Change to requested working directory
cd $firstpath

# Load first file
ReadFile $firstfile {}

# Initialize and show coords display
bind $coords_win <Configure> {}  ;# Not needed
WaitVisible $coords_win
InitCoords $coords_win $DisplayRotation
UpdateCoordDisplay $DisplayRotation
after idle UpdateCoordColors
## All the axes except the first are drawn through the
## after idle event loop.

# Release firstpath & firstfile variables
unset firstpath firstfile

focus $canvas_frame

# Initialize omfSavedDisplayState
OMF_SaveDisplayState

after idle Ow_SetIcon .
